#!/usr/bin/perl -w

use strict;
use Socket;
use Compress::Zlib;
#use Carp qw(cluck);

# from <netinet/in.h>, verified on Linux x86
use constant IPPROTO_IP => 0;
use constant IP_ADD_MEMBERSHIP => 35;

if (0) {

do 'AAAArrr4';
for my $i (0 .. $#main::pkts) {
#for my $i (qw(1314 2647)) {
#$my $i = 998; {
#	print '-' x 78 . "\n";
#	print "I: $i\n";
	my $sap = sap_parser($main::pkts[$i]) || next;
	foreach (qw(source id auth deletion type)) {
#		print uc($_) . ': ' . $sap->{$_} . "\n" if defined $sap->{$_};
	}
	if ($sap->{deletion}) {
#		print $sap->{payload} . "\n";
		next;
	}
	my $sdp = sdp_parser($sap->{payload});
#	print "===\n$data===\n";
	use Data::Dumper; print Data::Dumper->Dumpxs([$sdp], ['sdp']); #exit 0;
}
exit 0;

}


my $sock = open_multicast('224.2.127.254', 9875);

print "=======================================\n";
while (1) {
	my $data;
	my $addr = recv(SOCK, $data,  2048, 0) or warn "recv: $!";
	print "-----------------------------------\n";
	my $sap = sap_parser($data);
	if (not defined $sap) {
		my ($srcport, $srcip) = sockaddr_in($addr);
		my $srcaddr = inet_ntoa($srcip);
		print "BAD PACKET from $srcaddr:$srcport.\n";
		next;
	}
	if ($sap->{deletion}) {
		print "DELETION: " . $sap->{payload} . "\n"; # XXX
		next;
	}
	my $sdp = sdp_parser($sap->{payload});
	use Data::Dumper; print Data::Dumper->Dumpxs([$sdp], ['sdp']); #exit 0;
}

exit 0;

sub sap_parser {
	my ($data) = @_;

	my $header = substr($data, 0, 4, '');
	my ($flags, $auth_len, $msgid) = unpack('B8 C n', $header);
	my ($v1, $v2, $v3, $flag_a, $flag_r, $flag_t, $flag_e, $flag_c)
		= split(//, $flags);

	my $source;
	if ($flag_a) {	# IPv6
		$source = sprintf('%0*v2x', ':', substr($data, 0, 16, ''));
	} else {		# IPv4
		$source = join('.', unpack('C4', substr($data, 0, 4, '')));
	}
	my $auth = substr($data, 0, $auth_len * 4, '');

	if (not defined $source or not defined $msgid) {
		warn "BAD PACKET <$source> <$msgid>";
		return;
	}
	if ($flag_e) {
		warn "UNIMPLEMENTED: E:$flag_e";
#print "<<<<<<<<$data>>>>>>>>>>\n";
#use Data::Dumper; print Data::Dumper->Dumpxs([\%sdp], ['*sdp']); #exit 0;
		return;
	}

	# try to remove the MIME type. what a stupid format
	$data =~ s#^([a-zA-Z0-9\.-]+/[a-zA-Z0-9\.-]+)\0##;
	my $type = $1;

	# the "MGW 2000" streaming server adds a trailing NUL
	$data =~ s/\000$//;

	my %sap = (
		source	=> $source,
		payload	=> ($flag_c ? uncompress($data) : $data),
		id		=> $msgid,
	);
	$sap{type} = $type if $type;
	$sap{auth} = $auth if $auth;
	$sap{deletion} = 1 if $flag_t;

	return \%sap;
}

sub sdp_parser {
	my ($data) = @_;

	my @lines = split(/\n/, $data);

#	my @fields = qw(v o s i* u* e* p* c* b*);
	# (t r*)+ z* k* a* (m i* c* b* k* a*)*

	my %sdp;
	foreach my $ptype (qw(v o s i u e p c b)) {
		my $line = $lines[0];
		if (not $line) {
			warn;
			return undef;
		}

		# RealProducer Basic 8.5.0.391 Windows is a broken POS
		$line =~ s/^s=$/s=?/;

		my ($type, $value) = ($line =~ /^([a-z])=(.+)/);
		if (not $type or $type ne $ptype) {
			next unless $ptype =~ /^[vos]$/;
			warn "BROKEN SDP PACKET1 (missing $ptype): \n===\n$data\n==="
				if $ptype ne 'o';
			return undef;
		}

		$sdp{$type} = $value;
		shift @lines;
	}

	while (@lines and $lines[0] =~ /^t=(.+)/) {
		my %t = (t => $1);
		shift @lines;

		while (@lines and $lines[0] =~ /^r=(.+)/) {
			push(@{$t{r}}, $1);
			shift @lines;
		}

		push(@{$sdp{t}}, \%t);
	}

	foreach my $ptype (qw(z k)) {
		my $line = $lines[0];
		if (not $line) {
			warn;
			return undef;
		}
		my ($type, $value) = ($line =~ /^([a-z])=(.+)/);

		if (not $type) {
			warn "BROKEN SDP PACKET2 ($line): \n===\n$data\n===";
			return undef;
		}
		next if $type ne $ptype;

		$sdp{$type} = $value;
		shift @lines;
	}

	while (@lines and $lines[0] =~ /^a=(.+)/) {
		push(@{$sdp{a}}, $1);
		shift @lines;
	}

	while (@lines and $lines[0] =~ /^m=(.+)/) {
		my %m = (m => $1);
		shift @lines;

		foreach my $ptype (qw(i c b k)) {
			my $line = $lines[0] or last;
			if (not $line) {
				warn "BROKEN SDP PACKET4 (type: $ptype): \n===\n$data\n===";
				return undef;
			}
			my ($type, $value) = ($line =~ /^([a-z])=(.+)/);

			if (not $type) {
				warn "BROKEN SDP PACKET5 ($line): \n===\n$data\n===";
				return undef;
			}
			next if $type ne $ptype;

			$m{$type} = $value;
			shift @lines;
		}

		while (@lines and $lines[0] =~ /^a=(.+)/) {
			push(@{$m{a}}, $1);
			shift @lines;
		}

		push(@{$sdp{m}}, \%m);
	}

	if (@lines) {
		warn "BROKEN SDP PACKET6 (".join('|',@lines)."): \n===\n$data\n==="
			if $lines[0] !~ /^t/;
		return undef;
	}

	return \%sdp;
}

sub open_multicast {
	my ($group, $port) = @_;

	socket(SOCK, PF_INET, SOCK_DGRAM, 0) or die $!;

	bind(SOCK, sockaddr_in($port, inet_aton($group))) or die $!;

	my $ip_mreq = inet_aton($group) . inet_aton('0.0.0.0');
	setsockopt(SOCK, IPPROTO_IP, IP_ADD_MEMBERSHIP, $ip_mreq) or die $!;

	return \*SOCK;
}

